#include "Python.h"             /* Python header files */

#include "data_rows.h"

#include "DataRows.hlp"

#include "utility.h"

static PyObject *ErrorObject;   /* locally-raised exception */

#define RETURN_OBJ_ERROR(message) \
       { PyErr_SetString(ErrorObject, message); return NULL; }

#define RETURN_INT_ERROR(message) \
       { PyErr_SetString(ErrorObject, message); return -1; }

/*****************************************************************************
 * TYPE INFORMATION
 *****************************************************************************/

typedef struct
{
    PyObject_HEAD
    Data_rows *data_rows;
}   DataRows_object;

static PyTypeObject DataRows_type;

PyDoc_STRVAR(DataRows_doc, DataRows_help);

Bool is_data_rows_object(PyObject *obj)
{
/*  below does not work because different *.so files end up
    with different addresses for DataRows_type
    return (obj->ob_type == &DataRows_type);
*/
    return equal_strings((char *) obj->ob_type->tp_name, (char *) DataRows_type.tp_name);
}

/*****************************************************************************
 * MISCELLANEOUS METHODS
 *****************************************************************************/

static PyObject *get_int_list_object(int n, int *array)
{
    int i;
    PyObject *list, *z;

    list = PyList_New(n);

    for (i = 0; i < n; i++)
    {
	z = PyInt_FromLong((long) array[i]);

	PyList_SetItem(list, i, z);
    }

    return list;
}

/*****************************************************************************
 * INSTANCE METHODS
 *****************************************************************************/

static PyObject *data_rows_index_method(PyObject *self, PyObject *args)
{
    DataRows_object *obj = (DataRows_object *) self;
    Data_rows *d = obj->data_rows;
    PyObject *list, *z;
    int point[MAX_NDIM], i, n;
    Line error_msg;

    if (!PyArg_ParseTuple(args, "O", &list))
        RETURN_OBJ_ERROR("must have one argument, the point");

    if (!PyList_Check(list))
	RETURN_OBJ_ERROR("point must be a list");

    if (PyList_Size(list) != d->size_info.ndim)
    {
	sprintf(error_msg,
		"point must be a list of length ndim (%d)", d->size_info.ndim);
	RETURN_OBJ_ERROR(error_msg);
    }

    for (i = 0; i < d->size_info.ndim; i++)
    {
	z = PyList_GetItem(list, i);

	if (!PyInt_Check(z))
	    RETURN_OBJ_ERROR("data must be a list of integers");

	point[i] = (int) PyInt_AsLong(z);

	if ((i != d->dim) &&
		((point[i] <= 0) || (point[i] > d->size_info.npoints[i])))
	{
	    sprintf(error_msg,
		"point[%d] must be > 0 and <= %d", i, d->size_info.npoints[i]);
	    RETURN_OBJ_ERROR(error_msg);
	}

	point[i]--; /* point input starts counting at 1 not 0 */ 
    }

    point[d->dim] = 0; /* this should be set 0 but not required for input */

    n = data_rows_index(d, point);

    return Py_BuildValue("i", n);
}

static PyObject *data_rows_point_method(PyObject *self, PyObject *args)
{
    DataRows_object *obj = (DataRows_object *) self;
    Data_rows *d = obj->data_rows;
    int point[MAX_NDIM], i, n;
    Line error_msg;

    if (!PyArg_ParseTuple(args, "i", &n))
        RETURN_OBJ_ERROR("must have one argument, the index");

    if ((n < 0) || (n >= d->nrows))
    {
	sprintf(error_msg, "index must be >= 0 and < %d", d->nrows);
	RETURN_OBJ_ERROR(error_msg);
    }

    data_rows_point(d, n, point);

    for (i = 0; i < d->size_info.ndim; i++)
    {
	if (i != d->dim)
	    point[i]++; /* point output starts counting from 1 not 0 */
    }

    return get_int_list_object(d->size_info.ndim, point);
}

static struct PyMethodDef data_rows_methods[] =
{
    { "index",	data_rows_index_method,	METH_VARARGS },
    { "point",	data_rows_point_method,	METH_VARARGS },
    { NULL,	NULL,			0 }
};

/*****************************************************************************
 * BASIC TYPE-OPERATIONS
 *****************************************************************************/

static PyObject *new_data_rows_object(String name, int dim, String mode, int in_order)
{
    DataRows_object *obj;
    Data_rows *d;
    Line error_msg;

/*
    printf("in new_data_rows_object\n");
*/

    d = new_data_rows(name, dim, mode, in_order, error_msg);

    if (d == NULL)
	RETURN_OBJ_ERROR(error_msg);

    obj = PyObject_NEW(DataRows_object, &DataRows_type);

    if (obj == NULL)
    {
	delete_data_rows(d);

	RETURN_OBJ_ERROR("allocating data_rows object");
    }

    obj->data_rows = d;

    return  (PyObject *) obj;
}

static void delete_data_rows_object(PyObject *self)
{
    DataRows_object *obj = (DataRows_object *) self;
    Data_rows *d = obj->data_rows;

/*
    printf("in delete_data_rows_object\n");
*/

    delete_data_rows(d);

    PyObject_DEL(self);
}

/*
static int print_data_rows_object(PyObject *self, FILE *fp, int flags)
{
    printf("in print_data_rows_object\n");

    return 0;
}
*/

static PyObject *getattr_data_rows_object(PyObject *self, char *name)
{
    DataRows_object *obj = (DataRows_object *) self;
    Data_rows *d = obj->data_rows;

/*
    printf("in getattr_data_rows_object\n");
*/

    if (strcmp(name, "name") == 0)
	return Py_BuildValue("s", d->name);
    else if (strcmp(name, "dim") == 0)
	return Py_BuildValue("i", d->dim + 1);
    else if (strcmp(name, "ndim") == 0)
	return Py_BuildValue("i", d->size_info.ndim);
    else if (strcmp(name, "row_size") == 0)
	return Py_BuildValue("i", d->row_size);
    else if (strcmp(name, "nrows") == 0)
	return Py_BuildValue("i", d->nrows);
    else if (strcmp(name, "npoints") == 0)
	return get_int_list_object(d->size_info.ndim, d->size_info.npoints);
    else if (strcmp(name, "block_size") == 0)
	return get_int_list_object(d->size_info.ndim, d->size_info.block_size);
    else
	return Py_FindMethod(data_rows_methods, self, name);
}

/*
static int setattr_data_rows_object(PyObject *self, char *name, PyObject *value)
{
    printf("in setattr_data_rows_object\n");

    return 0;
}
*/

static Py_ssize_t data_rows_length_object(PyObject *self)
{
    DataRows_object *obj = (DataRows_object *) self;
    Data_rows *d = obj->data_rows;
    Py_ssize_t nrows = d->nrows;

/*
    printf("in data_rows_length_object\n");
*/

    return nrows;
}

/*
static PyObject *data_rows_concat_object(PyObject *obj1, PyObject *obj2)
{
    printf("in data_rows_concat_object\n");

    Py_INCREF(Py_None);

    return Py_None;
}
*/

/*
static PyObject *data_rows_repeat_object(PyObject *obj, int n)
{
    printf("in data_rows_repeat_object\n");

    Py_INCREF(Py_None);

    return Py_None;
}
*/

static void set_row_item(int n, float x, Generic_ptr data)
{
    PyObject *list = (PyObject *) data;
    PyObject *z = PyFloat_FromDouble((double) x);

    PyList_SetItem(list, n, z);
}

static PyObject *data_rows_item_object(PyObject *self, Py_ssize_t nn)
{
    DataRows_object *obj = (DataRows_object *) self;
    Data_rows *d = obj->data_rows;
    int n = nn;
    PyObject *list;
    Line error_msg;

/*
    printf("in data_rows_item_object\n");
*/

    if (d->write_only)
	RETURN_OBJ_ERROR("data is write only");

    if ((n < 0) || (n >= d->nrows))
    {
	sprintf(error_msg, "index must be >= 0 and < %d", d->nrows);
	RETURN_OBJ_ERROR(error_msg);
    }

    list = PyList_New(d->row_size);

    if (data_rows_get_row(d, n, set_row_item,
			(Generic_ptr) list, error_msg) == ERROR)
	RETURN_OBJ_ERROR(error_msg);

    return list;
}

/*
static PyObject *data_rows_slice_object(PyObject *obj, int low, int high)
{
    printf("in data_rows_slice_object\n");

    return obj;
}
*/

static float get_row_item(int n, Generic_ptr data)
{
    PyObject *list = (PyObject *) data;
    PyObject *z = PyList_GetItem(list, n);

    if (PyFloat_Check(z))
	return (float) PyFloat_AsDouble(z);
    else if (PyInt_Check(z))
	return (float) PyInt_AsLong(z);

    return 0.0; /* arbitrary, should never get here */
}

static int data_rows_ass_item_object(PyObject *self, Py_ssize_t nn, PyObject *list)
{
    DataRows_object *obj = (DataRows_object *) self;
    Data_rows *d = obj->data_rows;
    int i, n = nn;
    Line error_msg;
    PyObject *z;

/*
    printf("in data_rows_ass_item_object\n");
*/

    if (d->read_only)
	RETURN_INT_ERROR("data is read only");

    if ((n < 0) || (n >= d->nrows))
    {
	sprintf(error_msg, "index must be >= 0 and < %d", d->nrows);
	RETURN_INT_ERROR(error_msg);
    }

    if (!PyList_Check(list))
	RETURN_INT_ERROR("data must be a list");

    if (PyList_Size(list) != d->row_size)
    {
	sprintf(error_msg,
		"data must be a list of length row_size (%d)", d->row_size);
	RETURN_INT_ERROR(error_msg);
    }

    for (i = 0; i < d->row_size; i++)
    {
	z = PyList_GetItem(list, i);

	if (!PyFloat_Check(z) && !PyInt_Check(z))
	    RETURN_INT_ERROR("data must be a list of floats or integers");
    }

    if (data_rows_set_row(d, n, get_row_item,
			(Generic_ptr) list, error_msg) == ERROR)
	RETURN_INT_ERROR(error_msg);

    return 0;
}

/*
static int data_rows_ass_slice_object(PyObject *obj, int low, int high, PyObject *v)
{
    printf("in data_rows_ass_slice_object\n");

    return 0;
}
*/

/*****************************************************************************
 * TYPE DESCRIPTORS
 *****************************************************************************/

/*  if implementing more...
static PySequenceMethods DataRows_sequence_methods =
{
    data_rows_length_object,
    data_rows_concat_object,
    data_rows_repeat_object,
    data_rows_item_object,
    data_rows_slice_object,
    data_rows_ass_item_object,
    data_rows_ass_slice_object
};
*/

static PySequenceMethods DataRows_sequence_methods =
{
    data_rows_length_object,
    0,
    0,
    data_rows_item_object,
    0,
    data_rows_ass_item_object,
    0
};

static PyTypeObject DataRows_type =
{
#ifdef WINDOWS
    1, NULL,
#else
    PyObject_HEAD_INIT(&PyType_Type)
#endif
    0,
    "DataRows", /* name */
    sizeof(DataRows_object), /* basicsize */
    0, /* itemsize */
    delete_data_rows_object, /* destructor */
    0, /* printfunc */
    getattr_data_rows_object, /* getattr */
    0, /* setattr */
    0, /* cmpfunc */
    0, /* reprfunc */
    0, /* PyNumberMethods */
    &DataRows_sequence_methods /* PySequenceMethods */
};

/*****************************************************************************
 * MODULE LOGIC
 *****************************************************************************/

static PyObject *init_data_rows_object(PyObject *self, PyObject *args, PyObject *keywds)
{
    String name, mode;
    int dim, in_order = 0;
    static char *kwlist[] = { "par_file", "dim", "mode", "in_order", NULL };

    if (!PyArg_ParseTupleAndKeywords(args, keywds, "sis|i", kwlist, &name, &dim, &mode, &in_order))
        RETURN_OBJ_ERROR("arguments: par_file; dim; mode ('r' or 'w'); in_order (optional)");

    return  new_data_rows_object(name, dim, mode, in_order);
}

/******************************************************************************
* METHOD REGISTRATION TABLE: NAME-STRING -> FUNCTION-POINTER
*
* List of functions defined in the module. A name->address method map, used
* to build-up the module's dictionary in "Py_InitModule". Once imported, this
* module acts just like it's coded in Python. The method functions handle
* converting data from/to python objects, and linkage to other C functions.
******************************************************************************/


static struct PyMethodDef data_rows_type_methods[] =
{
    { "DataRows",	(PyCFunction) init_data_rows_object,	METH_VARARGS|METH_KEYWORDS },
    { NULL,		NULL,			0 }
};


/******************************************************************************
* INITIALIZATION FUNCTION (IMPORT-TIME)
*
* Initialization function for the module. Called on first "import DataRows" in 
* a Python program. The function is usually called "initDataRows": this name's
* added to the built-in module table in config.c statically (if added to file
* Module/Setup), or called when the module's loaded dynamically as a shareable 
* object-file found on PYTHONPATH. File and function names matter if dynamic.
******************************************************************************/

void
initDataRows()
{
    PyObject *m;

#ifdef WINDOWS
    DataRows_type.ob_type = &PyType_Type;
#endif

    /* create the module and add the functions */
    m = Py_InitModule3("DataRows", data_rows_type_methods, DataRows_doc);

    /* create exception object and add to module */
    ErrorObject = PyErr_NewException("DataRows.error", NULL, NULL);
    Py_INCREF(ErrorObject);
    PyModule_AddObject(m, "error", ErrorObject);

    /* check for errors */
    if (PyErr_Occurred())
        Py_FatalError("can't initialize module DataRows");
}
