#include "ansig.h"

#define  MAX_NDIM	10

static int word_count;

static int nindex_alloc = 0;
static int *index_table = (int *) NULL;

static int npolyline_alloc = 0;
static float *x = (float *) NULL;
static float *y = (float *) NULL;

static int nsub_planes;

static int start[MAX_NDIM];
static int cumul1[MAX_NDIM];
static int cumul2[MAX_NDIM];
static int array[MAX_NDIM];

static Status get_float(FILE *fp, float *x, String error_msg)
{
    if (FREAD(x, BYTES_PER_WORD, 1, fp))
    {
	sprintf(error_msg, "reading word %d", word_count);
	return  ERROR;
    }

    word_count++;

    return  OK;
}

static Status get_float_int(FILE *fp, int *d, String error_msg)
{
    float x;

    CHECK_STATUS(get_float(fp, &x, error_msg));

    *d = NEAREST_INTEGER(x);

    return  OK;
}

static Status seek_up_to_block_size(FILE *fp, int block_size, String error_msg)
{
    int r = word_count % block_size;

    if (r > 0)
    {
	r = block_size - r;

	if (FSEEK_RELATIVE(fp, r))
	    RETURN_ERROR_MSG("seeking over block");

	word_count += r;
    }

    return  OK;
}

static Status alloc_header_memory(Header_info *info, String error_msg)
{
    int ndim = info->ndim;

    sprintf(error_msg, "allocating header memory");

    MALLOC(info->dim, int, ndim);
    MALLOC(info->dim_map, int, ndim);
    MALLOC(info->lower, float, ndim);
    MALLOC(info->upper, float, ndim);
    MALLOC(info->pnts, int, ndim);

    return  OK;
}

static Status read_header(FILE *fp, Header_info *info, String error_msg)
{
    int i, dim, dim0, dim1, ndim, pnts;
    float lower, upper, ppm0, ppm1, version;

    word_count = 0;

    CHECK_STATUS(get_float(fp, &version, error_msg));

    info->version = version;

    CHECK_STATUS(get_float_int(fp, &ndim, error_msg));

    info->ndim = ndim;

    if (ndim < 2)
    {
	sprintf(error_msg, "ndim = %d, must be at least 2", ndim);
	return  ERROR;
    }

    CHECK_STATUS(alloc_header_memory(info, error_msg));

    CHECK_STATUS(get_float_int(fp, &dim1, error_msg));
    CHECK_STATUS(get_float_int(fp, &dim0, error_msg));

    dim0 = ndim - dim0;
    dim1 = ndim - dim1;

    info->dim[0] = dim0;
    info->dim[1] = dim1;

    info->dim_map[0] = dim1;
    info->dim_map[1] = dim0;

    for (i = 2; i < ndim; i++)
    {
	CHECK_STATUS(get_float_int(fp, &dim, error_msg));
	dim = ndim - dim;

	info->dim[ndim-i+1] = dim;

	info->dim_map[i] = dim;
    }

    for (i = 0; i < ndim; i++)
    {
	CHECK_STATUS(get_float(fp, &upper, error_msg));
	CHECK_STATUS(get_float(fp, &lower, error_msg));

	dim = info->dim_map[i];

	info->lower[dim] = lower;
	info->upper[dim] = upper;
    }

    CHECK_STATUS(get_float(fp, &ppm1, error_msg));
    CHECK_STATUS(get_float(fp, &ppm0, error_msg));

    info->ppm0 = ppm0;
    info->ppm1 = ppm1;

    for (i = 0; i < ndim; i++)
    {
	CHECK_STATUS(get_float_int(fp, &pnts, error_msg));

	dim = info->dim_map[i];

	info->pnts[dim] = pnts;
    }

    SWAP(info->dim_map[0], info->dim_map[1], int);

    return  OK;
}

Status read_contours_header(String file, Header_info *info, String error_msg)
{
    FILE *fp;
    Status status;

    CHECK_OPEN_FOR_READING(fp, file);

    status = read_header(fp, info, error_msg);

    fclose(fp);

    return  status;
}

static void free_index_table()
{
    FREE(index_table, int);

    nindex_alloc = 0;
}

static Status alloc_index_table(int n, String error_msg)
{
    if (n > nindex_alloc)
    {
	free_index_table();

	sprintf(error_msg, "allocating index table");

	MALLOC(index_table, int, n);

	nindex_alloc = n;
    }

    return  OK;
}

Status read_index_table(FILE *fp, Header_info *info, int block_size,
							String error_msg)
{
    int p, r;

    VECTOR_PRODUCT(p, info->pnts, info->ndim);

    p *= 2;

    CHECK_STATUS(alloc_index_table(p, error_msg));

    if (FREAD(index_table, BYTES_PER_WORD, p, fp))
	RETURN_ERROR_MSG("reading index table");

    word_count += p;

    CHECK_STATUS(seek_up_to_block_size(fp, block_size, error_msg));

    return  OK;
}

static int determine_sub_planes(Header_info *info, float *lower, float *upper)
{
    int i, n, n0, n1, m, nsub_planes, dim;
    float l, u, d;

    nsub_planes = 1;
    m = 1;

    for (i = 0; i < info->ndim; i++)
    {
	dim = info->dim[i];

	l = lower[dim] - info->lower[dim];
	u = upper[dim] - info->lower[dim];
	d = info->upper[dim] - info->lower[dim];
	n = info->pnts[dim];

	if (d == 0)
	{
	    n0 = 0;
	    n1 = 1;
	}
	else if (i < 2)
	{
	    n0 = FLOOR(n * l / d);
	    n1 = CEILING(n * u / d);
	}
	else
	{
	    n0 = CEILING((n-1) * l / d);
	    n1 = 1 + FLOOR((n-1) * u / d);
	}

	n0 = MAX(n0, 0);
	n1 = MAX(n1, 0);
	n0 = MIN(n0, n);
	n1 = MIN(n1, n);

	start[dim] = n0;
	cumul1[dim] = nsub_planes;
	cumul2[dim] = m;

	nsub_planes *= n1 - n0;
	m *= n;
    }

    return  nsub_planes;
}

static int determine_sub_plane(int n, int ndim)
{
    int p;

    ARRAY_OF_INDEX(array, n, cumul1, ndim);
    ADD_VECTORS(array, array, start, ndim);
    INDEX_OF_ARRAY(p, array, cumul2, ndim);

    return  p;
}

static void free_polyline_memory()
{
    FREE(x, float);
    FREE(y, float);

    npolyline_alloc = 0;
}

static Status alloc_polyline_memory(int n, String error_msg)
{
    if (n > npolyline_alloc)
    {
	free_polyline_memory();

	sprintf(error_msg, "allocating polyline memory");

	MALLOC(x, float, n);
	MALLOC(y, float, n);

	npolyline_alloc = n;
    }

    return  OK;
}

static Status read_polyline(int n, FILE *fp, Contour_polyline_func func,
					int pos_neg, String error_msg)
{
    int i;

    CHECK_STATUS(alloc_polyline_memory(n, error_msg));

    for (i = 0; i < n; i++)
    {
	CHECK_STATUS(get_float(fp, y+i, error_msg));
	CHECK_STATUS(get_float(fp, x+i, error_msg));
    }

    CHECK_STATUS((*func)(pos_neg, n, x, y, error_msg));

    return  OK;
}

static Status read_polylines(FILE *fp, int block_size, int offset,
		Contour_polyline_func func, int pos_neg, String error_msg)
{
    int o, n;

    offset--;

    o = offset - word_count;

    if (FSEEK_RELATIVE(fp, o))
	RETURN_ERROR_MSG("seeking data");

    word_count = offset;

    CHECK_STATUS(get_float_int(fp, &n, error_msg));

    while (n >= 0)
    {
	if (n == 0)
	{
	    CHECK_STATUS(seek_up_to_block_size(fp, block_size, error_msg));
	    CHECK_STATUS(get_float_int(fp, &n, error_msg));
	}

	CHECK_STATUS(read_polyline(n, fp, func, pos_neg, error_msg));
	CHECK_STATUS(get_float_int(fp, &n, error_msg));
    }


    return  OK;
}

static Status read_data(FILE *fp, Header_info *info, int block_size,
			float *lower, float *upper,
			Contour_polyline_func func, String error_msg)
{
    int i, h, n, offset, nsub_planes;

    h = 4*info->ndim + 4;

    if (FSEEK_RELATIVE(fp, h))
	RETURN_ERROR_MSG("seeking over header");

    CHECK_STATUS(read_index_table(fp, info, block_size, error_msg));

    nsub_planes = determine_sub_planes(info, lower, upper);

    for (i = 0; i < nsub_planes; i++)
    {
	n = determine_sub_plane(i, info->ndim);

	offset = index_table[2*n];  /* negative contours */

	if (offset != 0)
	    CHECK_STATUS(read_polylines(fp, block_size, offset, func, NEGATIVE_CONTOUR, error_msg));

	offset = index_table[2*n+1];  /* positive contours */

	if (offset != 0)
	    CHECK_STATUS(read_polylines(fp, block_size, offset, func, POSITIVE_CONTOUR, error_msg));
    }

    return  OK;
}

Status read_contours_data(String file, Header_info *info,
				int block_size, float *lower, float *upper,
				Contour_polyline_func func, String error_msg)
{
    int i;

    FILE *fp;
    Status status;

    if (info->ndim > MAX_NDIM)
	RETURN_ERROR_MSG("ndim too large");

    for (i = 0; i < info->ndim; i++)
    {
	if (lower[i] > upper[i])
	{
	    sprintf(error_msg, "%d: lower (%f) > upper (%f)",
						i, lower[i], upper[i]);
	    return  ERROR;
	}
    }

    CHECK_OPEN_FOR_READING(fp, file);

    status = read_data(fp, info, block_size, lower, upper, func, error_msg);

    fclose(fp);

    return  status;
}
