#include "subset.h"

#include "par.h"
#include "utility.h"

static int ndim;
static int *npoints_in;
static int *block_size;

static int size_alloc = 0;
static float *store_in = NULL;
static float *store_out;
static float *s_in;
static float *s_out;

static int *first;
static int *last;

static Bool swapped;
static Bool integer;
static Bool blocked;
static int header;
static FILE *file;

static int total_nblocks;
static int size_of_block;
static int disk_block;
static int base_block;
static int npts;

static int npoints_out[MAX_NDIM];
static int nblocks_in[MAX_NDIM];
static int nblocks_out[MAX_NDIM];
static int array[MAX_NDIM];
static int cum_nblocks_in[MAX_NDIM];
static int cum_nblocks_out[MAX_NDIM];
static int cum_npoints_out[MAX_NDIM];
static int cum_block_size[MAX_NDIM];
static int cum_npts[MAX_NDIM];
static int base_pnt_in[MAX_NDIM];
static int base_pnt_out[MAX_NDIM];
static int first_block[MAX_NDIM];

static void init_arrays()
{
    int i, n;

    BLOCKS(nblocks_in, npoints_in, block_size, ndim);

    for (i = 0; i < ndim; i++)
    {
	first_block[i] = first[i] / block_size[i];
	nblocks_out[i] = (last[i]-1)/block_size[i] - first_block[i] + 1;
	npoints_out[i] = last[i] - first[i];
    }

    CUMULATIVE(cum_nblocks_in, nblocks_in, n, ndim);
    CUMULATIVE(cum_nblocks_out, nblocks_out, total_nblocks, ndim);
    CUMULATIVE(cum_block_size, block_size, size_of_block, ndim);
    CUMULATIVE(cum_npoints_out, npoints_out, n, ndim);

    INDEX_OF_ARRAY(base_block, first_block, cum_nblocks_in, ndim);
}

static Status alloc_store()
{
    if (size_of_block > size_alloc)
    {
	FREE(store_in, float);
	size_alloc = 0;
	MALLOC(store_in, float, size_of_block);
	size_alloc = size_of_block;
    }

    return  OK;
}

static Status read_block(int block, String error_msg)
{
    long int delta_block;

    delta_block = size_of_block * (block - disk_block);

    if (FSEEK_RELATIVE(file, delta_block))
    {
        sprintf(error_msg, "seeking block %d in read_block", block);
        return  ERROR;
    }

    if (FREAD(store_in, BYTES_PER_WORD, size_of_block, file))
    {
	sprintf(error_msg, "reading block %d in read_block", block);
	return  ERROR;
    }

    if (swapped)
	swap_bytes((char *) store_in, BYTES_PER_WORD*size_of_block);

    if (integer)
	float_words(store_in, size_of_block);

    disk_block = block + 1;

    return  OK;
}

static Status init_block(int block, String error_msg)
{
    int i, f, n;

    ARRAY_OF_INDEX(array, block, cum_nblocks_out, ndim);
    INDEX_OF_ARRAY(block, array, cum_nblocks_in, ndim);
    CHECK_STATUS(read_block(block, error_msg));

    npts = 1;
    for (i = 0; i < ndim; i++)
    {
	f = first[i] % block_size[i];

	if (array[i] == 0)
	{
	    base_pnt_in[i] = f;
	    base_pnt_out[i] = 0;

	    if (array[i] == (nblocks_out[i]-1))
		n = npoints_out[i];
	    else
		n = block_size[i] - f;
	}
	else if (array[i] == (nblocks_out[i]-1))
	{
	    base_pnt_in[i] = 0;
	    base_pnt_out[i] = array[i]*block_size[i] - f;
	    n = 1 + ((last[i]-1) % block_size[i]);
	}
	else
	{
	    base_pnt_in[i] = 0;
	    base_pnt_out[i] = array[i]*block_size[i] - f;
	    n = block_size[i];
	}

	cum_npts[i] = npts;
	npts *= n;

    }

    INDEX_OF_ARRAY(n, base_pnt_in, cum_block_size, ndim);
    s_in = store_in + n;

    INDEX_OF_ARRAY(n, base_pnt_out, cum_npoints_out, ndim);
    s_out = store_out + n;

    return  OK;
}

static void extract_points()
{
    int i, pt_in, pt_out;

    for (i = 0; i < npts; i++)
    {
	ARRAY_OF_INDEX(array, i, cum_npts, ndim);

	INDEX_OF_ARRAY(pt_in, array, cum_block_size, ndim);
	INDEX_OF_ARRAY(pt_out, array, cum_npoints_out, ndim);

	s_out[pt_out] = s_in[pt_in];
    }
}

static Status block_process(String error_msg)
{
    int i;

    if (FSEEK_ABSOLUTE(file, header))  /* skip over header */
    {
	sprintf(error_msg, "seeking over input header in block_process");
	return  ERROR;
    }

    if (FSEEK_RELATIVE(file, size_of_block*base_block))
    {
        sprintf(error_msg, "seeking over first blocks in block_process");
        return  ERROR;
    }

    disk_block = 0;

    for (i = 0; i < total_nblocks; i++)
    {
	CHECK_STATUS(init_block(i, error_msg));
	extract_points();
    }

    return  OK;
}

Status extract_subset(Par_info *par_info, Subset_info *subset_info,
							String error_msg)
{
    String name;
    Status status;

    store_out = subset_info->store;
    first = subset_info->first;
    last = subset_info->last;

    ndim = par_info->ndim;
    npoints_in = par_info->npoints;
    block_size = par_info->block_size;
    name = par_info->file;
    swapped = par_info->swapped;
    integer = par_info->integer;
    blocked = par_info->blocked;
    header = par_info->header;

    if (!blocked)
	RETURN_ERROR_MSG("input file must be blocked in extract_subset");

    init_arrays();

    if (alloc_store() == ERROR)
	RETURN_ERROR_MSG("allocating memory for store in extract_subset");

    if (OPEN_FOR_BINARY_READING(file, name))
	RETURN_ERROR_MSG("opening input file in extract_subset");

    status = block_process(error_msg);

    fclose(file);

    return  status;
}
