#include "par.h"
#include "ref.h"
#include "utility.h"

#define  MAX_NDIM  10

#define XEASY_KEY_LEN  32

/* conversion routines from pipe2xeasy code by Ralf W. Glaser */

static float reverse[96];

static float xexp10(float x)
{
  return (float)(exp(x*log(10.0)));
}

static void init_reverse()
{
    int i;

    for(i = 1; i <= 47; i++)
	reverse[i] = (xexp10((float)(i - 1)/(float)6.6438562)+(float)0.5);
    for(i = 48; i <= 95; i++)
	reverse[i] = -(xexp10((float)(95 - i)/(float)6.6438562)+(float)0.5);
    reverse[0] = 1;
}

static void floattox16(float *realin, unsigned char* lo16)
{
    float real = *realin;
    int sign, lo, hi;

    if (real < 0)
    {
	sign = -1;
	real = -real;
    }
    else
    {
	sign = 1;
    }

    if (real < 0.5)
    {
	*lo16 = (unsigned char)0;
	*(lo16+1) = (unsigned char)0;
	return;
    }

    hi = (int)(log(real)*(float)2.8853902+1.414);
    if (hi < 0)
	hi = 0;

    if (hi > 47)
	hi = 47;

    lo = (int)(712*real/reverse[hi]-615);

    if (lo < 0)
	lo = 0;

    if (lo > 255)
	lo = 255;

    if ((sign<0) && hi)
	hi = 96-hi;

    *lo16 = (unsigned char)lo;
    *(lo16+1) = (unsigned char)hi;
}

static void printKey(FILE *fp, char *key)
{
    int i, ndots = XEASY_KEY_LEN - strlen(key) - 2;

    fprintf(fp, key);
    fprintf(fp, " ");
    for (i = 0; i < ndots; i++)
	fprintf(fp, ".");
    fprintf(fp, " ");
}

static void printKeyInt(FILE *fp, char *key, int value)
{
    printKey(fp, key);
    fprintf(fp, "%d\n", value);
}

static void printKeyFloat(FILE *fp, char *key, float value)
{
    printKey(fp, key);
    fprintf(fp, "%f\n", value);
}

static void printKeyString(FILE *fp, char *key, char *value)
{
    printKey(fp, key);
    fprintf(fp, "%s\n", value);
}

int main(int argc, char **argv)
{
    int i, ndim, p, block_size[MAX_NDIM], pts[MAX_NDIM];
    int xi;
    long n, npts, nprint;
    unsigned char *xc = (unsigned char *) &xi;
    float factor, x, v, sw;
    Line file_in, file_out, error_msg, line;
    Par_info par_info;
    FILE *fpr, *fpw;

    if (argc != 3 && argc != 4)
    {
	sprintf(error_msg,
	    "correct syntax: azara2xeasy <input par file> <output xeasy file prefix> [factor]");
	ERROR_AND_EXIT(error_msg);
    }

    strcpy(file_in, argv[1]);
    strcpy(file_out, argv[2]);
    strcat(file_out, ".3D.16");

    if (argc == 4)
	factor = atof(argv[3]);
    else
	factor = 1.0;

    if (read_par_file(file_in, &par_info, error_msg) == ERROR)
	ERROR_AND_EXIT(error_msg);

    if (par_info.deflated)
	ERROR_AND_EXIT("cannot convert deflated data");

    ndim = par_info.ndim;
    npts = 1;

    for (i = 0; i < ndim; i++)
    {
	if (par_info.blocked)
	    block_size[i] = par_info.block_size[i];
	else if (i == 0)
	    block_size[i] = par_info.npoints[i];
	else
	    block_size[i] = 1;

	p = par_info.npoints[i]; 
	if (p % block_size[i])
	    p += block_size[i] - (p % block_size[i]);

	pts[i] = p;
	npts *= p;
    }

    if ((fpr = fopen(par_info.file, BINARY_READ)) == NULL)
	ERROR_AND_EXIT("opening azara data file");

    if ((fpw = fopen(file_out, BINARY_WRITE)) == NULL)
	ERROR_AND_EXIT("opening xeasy data file");

    nprint = npts / 32;

    for (n = 0; fread((void *) &x, BYTES_PER_WORD, 1, fpr) > 0; n++)
    {
	if (!(n % nprint))
	    printf("Working on point %d of %d\n", n, npts);

	if (par_info.swapped)
	    swap_bytes((char *) &x, BYTES_PER_WORD);

	if (par_info.integer)
	    float_words(&x, 1);

	floattox16(&x, xc);

	if (fwrite((void *) &xi, 2, 1, fpw) != 1)
	    ERROR_AND_EXIT("writing UCSF data");
    }

    fclose(fpr);
    fclose(fpw);

    if (n != npts)
	printf("Uh oh, n = %d but npts = %d\n", n, npts);

    strcpy(file_out, argv[2]);
    strcat(file_out, ".3D.param");

    if ((fpw = fopen(file_out, WRITE)) == NULL)
	ERROR_AND_EXIT("opening xeasy param file");

    printKeyInt(fpw, "Version", 1);
    printKeyInt(fpw, "Number of dimensions", ndim);
    printKeyInt(fpw, "16 or 8 bit file type", 16);

    for (i = 0; i < ndim; i++)
    {
	sprintf(line, "Spectrometer frequency in w%d", i+1);
	printKeyFloat(fpw, line, par_info.ref[i].sf);
    }

    for (i = 0; i < ndim; i++)
    {
	sprintf(line, "Spectral sweep width in w%d", i+1);
	sw = (par_info.ref[i].sw * pts[i]) / par_info.npoints[i];
	printKeyFloat(fpw, line, sw/par_info.ref[i].sf);
    }

    for (i = 0; i < ndim; i++)
    {
	v = 1.0;
	convert_from_point(REF_PPM, par_info.npoints[i], par_info.ref+i, &v);
	sprintf(line, "Maximum chemical shift in w%d", i+1);
	printKeyFloat(fpw, line, v);
    }

    for (i = 0; i < ndim; i++)
    {
	sprintf(line, "Size of spectrum in w%d", i+1);
	p = par_info.npoints[i]; 
	if (p % block_size[i])
	    p += block_size[i] - (p % block_size[i]);
	printKeyInt(fpw, line, p);
    }

    for (i = 0; i < ndim; i++)
    {
	sprintf(line, "Submatrix size in w%d", i+1);
	printKeyInt(fpw, line, block_size[i]);
    }

    for (i = 0; i < ndim; i++)
    {
	sprintf(line, "Permutation for w%d", i+1);
	printKeyInt(fpw, line, i+1);
    }

    for (i = 0; i < ndim; i++)
    {
	sprintf(line, "Folding in w%d", i+1);
	printKeyString(fpw, line, "NO");
    }

    printKeyString(fpw, "Type of spectrum", "-");

    return 0;
}
