#include "macros.h"
#include "types.h"
#include "par.h"
#include "ref.h"

#define  OUTPUT_FILE_SUFFIX1	"_dist.bin"
#define  OUTPUT_FILE_SUFFIX2	"_sigma.bin"

#define  PAR_SUFFIX		".par"

#define  BIG_LINE_SIZE		5000

#define  SEPARATOR		" \t"

static Line input_file;
static Line output_file1;
static Line output_file2;
static Line prefix;
static FILE *fp_in;
static FILE *fp_out1;
static FILE *fp_out2;
static int nrows;
static int ncolumns;
static float **dist;
static float **sigma;
static float *row_data;
static char big_line[BIG_LINE_SIZE];

static Status open_files(String error_msg)
{
    CHECK_OPEN_FOR_READING(fp_in, input_file);
    CHECK_OPEN_FOR_WRITING(fp_out1, output_file1);
    CHECK_OPEN_FOR_WRITING(fp_out2, output_file2);

    return  OK;
}

static Status alloc_memory(String error_msg)
{
    int i;

    sprintf(error_msg, "allocating memory");

    MALLOC(dist, float *, nrows);
    MALLOC(sigma, float *, nrows);

    for (i = 0; i < nrows; i++)
    {
	MALLOC(dist[i], float, ncolumns);
	MALLOC(sigma[i], float, ncolumns);
    }

    MALLOC(row_data, float, ncolumns);

    return  OK;
}

static Status read_atom(String error_msg)
{
    Line atom;

    if (!fgets(atom, 20, fp_in))
	RETURN_ERROR_MSG("could not read atom");

    return  OK;
}

static Status read_line(String error_msg)
{
    if (!fgets(big_line, BIG_LINE_SIZE, fp_in))
	RETURN_ERROR_MSG("reading line");

    return  OK;
}

static Status read_row(String error_msg)
{
    int i;
    char *ptr1, *ptr2;

    CHECK_STATUS(read_atom(error_msg));
    CHECK_STATUS(read_line(error_msg));

    ptr1 = big_line;

    for (i = 0; i < ncolumns; i++)
    {
	if (!(ptr2 = strtok(ptr1, SEPARATOR)) ||
		(sscanf(ptr2, "%f", row_data + i) != 1))
	{
	    sprintf(error_msg, "reading column %d data", i);
	    return  ERROR;
	}

	ptr1 = NULL;
    }

    return  OK;
}

static Status count_number_rows(String error_msg)
{
    for (nrows = 0; read_line(error_msg) == OK; nrows++)
	;

    if (nrows == 0)
	RETURN_ERROR_MSG("no rows found");

    return  OK;
}

static Status count_number_columns(String error_msg)
{
    char *ptr;

    CHECK_STATUS(read_atom(error_msg));
    CHECK_STATUS(read_line(error_msg));

    ptr = big_line;
    for (ncolumns = 0; strtok(ptr, SEPARATOR); ncolumns++)
	ptr = NULL;

    if (ncolumns == 0)
	RETURN_ERROR_MSG("no columns found");

    return  OK;
}

static Status write_par_files(String error_msg)
{
    Par_info par_info;
    int npoints[2], block_size[2];
    Ref_info *ref = NULL;
    Line file;

    CHECK_STATUS(initialize_ref(2, &ref, error_msg));

    par_info.ndim = 2;
    par_info.npoints = npoints;
    par_info.ref = ref;
    par_info.block_size = block_size;
    par_info.swapped = FALSE;
    par_info.integer = FALSE;
    par_info.blocked = TRUE;
    par_info.header = 0;
    par_info.deflated = FALSE;
    par_info.level = 0;
    par_info.dir_size = 0;
    par_info.param_dim = -1;

    npoints[0] = ncolumns;
    npoints[1] = nrows;
    block_size[0] = ncolumns;
    block_size[1] = 1;

    sprintf(file, "%s%s", output_file1, PAR_SUFFIX);
    par_info.file = output_file1;
    CHECK_STATUS(write_par_file(file, &par_info, error_msg));

    sprintf(file, "%s%s", output_file2, PAR_SUFFIX);
    par_info.file = output_file2;
    CHECK_STATUS(write_par_file(file, &par_info, error_msg));

    return  OK;
}

static Status process_file(String error_msg)
{
    int i, j;

    CHECK_STATUS(open_files(error_msg));

    CHECK_STATUS(count_number_rows(error_msg));
    rewind(fp_in);

    CHECK_STATUS(count_number_columns(error_msg));
    rewind(fp_in);

    printf("found %d rows, %d columns\n", nrows, ncolumns);

    CHECK_STATUS(alloc_memory(error_msg));

    for (i = 0; i < nrows; i++)
    {
	sprintf(error_msg, "row %d: ", i);

	CHECK_STATUS(read_row(error_msg+strlen(error_msg)));

	for (j = 0; j < ncolumns; j++)
	{
	    if (i == j)
	    {
		dist[i][j] = sigma[i][j] = 0;
	    }
	    else if (i < j)
	    {
		dist[i][j] = dist[j][i] = row_data[j];
	    }
	    else /* (i > j) */
	    {
		sigma[i][j] = sigma[j][i] = row_data[j];
	    }
	}
    }

    fclose(fp_in);

    for (i = 0; i < nrows; i++)
    {
	if (FWRITE(dist[i], BYTES_PER_WORD, ncolumns, fp_out1))
	{
	    sprintf(error_msg, "writing dist data for row %d", i);
	    return  ERROR;
	}
    }

    fclose(fp_out1);

    for (i = 0; i < nrows; i++)
    {
	if (FWRITE(sigma[i], BYTES_PER_WORD, ncolumns, fp_out2))
	{
	    sprintf(error_msg, "writing sigma data for row %d", i);
	    return  ERROR;
	}
    }

    fclose(fp_out2);

    CHECK_STATUS(write_par_files(error_msg));

    return  OK;
}

int main(int argc, char **argv)
{
    char *ptr;
    Line error_msg;

    if ((argc != 2) && (argc != 3))
    {
	sprintf(error_msg,
	    "correct syntax: %s <input file> [ <output prefix> ]", argv[0]);
	ERROR_AND_EXIT(error_msg);
    }

    strcpy(input_file, argv[1]);

    if (argc == 2)
    {
	strcpy(prefix, input_file);

	if (ptr = strrchr(prefix, '.'))
	    *ptr = 0;
    }
    else
    {
	strcpy(prefix, argv[2]);
    }

    sprintf(output_file1, "%s%s", prefix, OUTPUT_FILE_SUFFIX1);
    sprintf(output_file2, "%s%s", prefix, OUTPUT_FILE_SUFFIX2);

    if (process_file(error_msg) == ERROR)
	ERROR_AND_EXIT(error_msg);

    return 0;
}
