#include "ref.h"
#include "utility.h"

#define  MAX_NDIM  10
#define  UCSF_FILE_HEADER  180
#define  UCSF_DIM_HEADER  128
#define  UCSF_MAX_HEADER  MAX(UCSF_FILE_HEADER, UCSF_DIM_HEADER)

#define  POINTS_READ  8192

static int get_int(int *x, Bool big)
{
    if (!big)
        swap_bytes((char *) x, BYTES_PER_WORD);

    return *x;
}
 
static float get_float(float *x, Bool big)
{
    if (!big)
        swap_bytes((char *) x, BYTES_PER_WORD);

    return *x;
}

int main(int argc, char **argv)
{
    int i, j, k, n, ndim, npoints[MAX_NDIM], block_size[MAX_NDIM], m = POINTS_READ;
    Bool big = is_big_endian();
    float x[POINTS_READ];
    char h[UCSF_MAX_HEADER], *f;
    Ref_info ref[MAX_NDIM];
    Line file_in, file_out, par_file, error_msg;
    FILE *fpr, *fpw;

    if (argc < 2)
    {
        sprintf(error_msg,
            "correct syntax: ucsf2azara <input UCSF file> [<output [par] file>]");
        ERROR_AND_EXIT(error_msg);
    }
 
    strcpy(file_in, argv[1]);
    if (argc >= 3)
    {
	f = argv[2];
	n = strlen(f);
        if ((n >= 4) && !strcmp(f+n-4, ".par"))
	{
	    *file_out = 0;
	    strcpy(par_file, f);
	}
	else
	{
	    strcpy(file_out, f);
	    sprintf(par_file, "%s.par", file_out);
	}
    }
    else
    {
	*file_out = 0;
	sprintf(par_file, "%s.par", file_in);
    }

    if (*file_out)
	printf("Copying data without header to '%s'\n", file_out);
    else
	printf("Not copying data just creating par file\n");
    printf("Par file = '%s'\n", par_file);

    if ((fpr = fopen(file_in, READ)) == NULL)
        ERROR_AND_EXIT("opening UCSF file");

    n = UCSF_FILE_HEADER;
    if (fread((void *) h, 1, n, fpr) != n)
	ERROR_AND_EXIT("reading UCSF file header");

    if (strncmp(h, "UCSF NMR", 8))
	ERROR_AND_EXIT("UCSF header must start with 'UCSF NMR'");

    ndim = (int) h[10];

    if (ndim <= 0)
    {
	sprintf(error_msg, "UCSF file has ndim = %d <= 0", ndim);
	ERROR_AND_EXIT(error_msg);
    }

    if (ndim > MAX_NDIM)
    {
	sprintf(error_msg, "UCSF file has ndim = %d, maximum is %d",
							ndim, MAX_NDIM);
	ERROR_AND_EXIT(error_msg);
    }

    n = UCSF_DIM_HEADER;
    for (i = 0; i < ndim; i++)
    {
	j = ndim - i - 1; /* UCSF does dims backwards */

	if (fread((void *) h, 1, n, fpr) != n)
	{
	    sprintf(error_msg, "reading UCSF header dim %d\n", i+1);
	    ERROR_AND_EXIT(error_msg);
	}

	strncpy(ref[j].nuc, h, 6);
	npoints[j] = get_int((int *) (h+8), big);
	block_size[j] = get_int((int *) (h+16), big);
	ref[j].sf = get_float((float *) (h+20), big);
	ref[j].sw = get_float((float *) (h+24), big);
	ref[j].refppm = get_float((float *) (h+28), big);
	ref[j].refpt = 0.5 * ((float) npoints[j]); /* center point */
    }

    if (*file_out)
    {
	if ((fpw = fopen(file_out, WRITE)) == NULL)
	    ERROR_AND_EXIT("opening output file");

	k = fread((void *) x, BYTES_PER_WORD, m, fpr);
	while (k > 0)
	{
	    if (fwrite((void *) x, BYTES_PER_WORD, k, fpw) != k)
		ERROR_AND_EXIT("writing UCSF data");
 
	    k = fread((void *) x, BYTES_PER_WORD, m, fpr);
	}

	fclose(fpw);
    }

    fclose(fpr);

    if ((fpr = fopen(par_file, WRITE)) == NULL)
        ERROR_AND_EXIT("opening par file for writing");
 
    fprintf(fpr, "! %s\n", par_file);
    fprintf(fpr, "\n");

    fprintf(fpr, "ndim %d\n", ndim);
    if (*file_out)
    {
	fprintf(fpr, "file %s\n", file_out);
    }
    else
    {
	fprintf(fpr, "file %s\n", file_in);
	n = (UCSF_FILE_HEADER + ndim*UCSF_DIM_HEADER) / 4;
	fprintf(fpr, "head %d\n", n);
    }
 
    if (!big)
        fprintf(fpr, "swap\n");
 
    for (i = 0; i < ndim; i++)
    {
        fprintf(fpr, "\n");
        fprintf(fpr, "dim %d\n", i+1);
        fprintf(fpr, "npts %d\n", npoints[i]);
        fprintf(fpr, "block %d\n", block_size[i]);
        fprintf(fpr, "sw %3.2f\n", ref[i].sw);
        fprintf(fpr, "sf %3.2f\n", ref[i].sf);
        fprintf(fpr, "refppm %3.2f\n", ref[i].refppm);
        fprintf(fpr, "refpt %3.2f\n", ref[i].refpt);
    }
 
    fclose(fpr);

    return 0;
}
